/* $Id: socket_database.c,v 1.31 2004/03/30 20:56:49 geschke Exp $ */
/* $Revision: 1.31 $ */
/* 
** This routines are for access to the database, either MySQL or Postgresql.
** Some parts are based on an early version of spo_database.c of the snort
** project. So probably
**
** Portions Copyright (C) 2000,2001 Carnegie Mellon University
** Portions Copyright (C) 2001 Jed Pickel <jed@pickel.net>
**
** and the remaining parts are
**
** Copyright (C) 2002-2004 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

static char *rcsid="$Id: socket_database.c,v 1.31 2004/03/30 20:56:49 geschke Exp $";

#include "socket_database.h"

#define TCP_OFFSET_DG(tcph)        (((tcph).th_offx2 & 0xf0) >> 4)
#define TCP_X2_DG(tcph)    ((tcph).th_offx2 & 0x0f)
#define IP_VER_DG(iph)     (((iph).ip_verhl & 0xf0) >> 4)
#define IP_HLEN_DG(iph)    ((iph).ip_verhl & 0x0f)

#ifdef GENUA
#warning GENUA defined!
#endif

#ifdef DEBUG
#define DEBUG_DB
#endif

/* in util.c: we first try to allocate as much memory as */
/* possible on startup. Thus we omit later checking for */
/* free memory */
void util_alloc();

static char *select_ptr=NULL; /* Memory is allocted in DatabaseInit! */
static char *insert_ptr=NULL; /* Memory is allocted in DatabaseInit! */
static char *variable=NULL; /* Needed in Database() */
static char *value=NULL;    /* Needed in Database() */
static char *tbuffer=NULL;  /* Needed in Database() to store palyoad */
static char *pbuffer=NULL;  /* Needed in Database() to store palyoad */

extern u_int LogSyslog;

/*
 * Function: DatabaseInit(DatabaseData *)
 *
 * Purpose: Connects to database, sets/gets the sensor id/name/encoding
 *          and selects the starting cid from database. Should we change
 *          this to scheme v106?
 *
 * Arguments: args => ptr to database parameters
 *
 * Returns: 0 on success
 *
 */

int 
DatabaseInit(DatabaseData *data)
{
  u_int32_t cid;
  int result, DBsuccess=0;

  /* This should never happen! */
  if(!data->sensor_name)
    {
      Printf("Database> ERROR: There is no sensor_name defined\n");
      exit(1);
    }

  data->tz = GetLocalTimeZone();

  /* allocate memory for configuration queries */
  if (select_ptr==NULL)
    {
      select_ptr = (char *) malloc(MAX_QUERY_LENGTH+1);
      insert_ptr = (char *) malloc(MAX_QUERY_LENGTH+1);
    }
  if ( (select_ptr==NULL) || (insert_ptr==NULL) )
	{
	  Printf("DatabaseInit: problem to allocate memory -> Exiting.\n");
	  return(1);
	}

  /* These values are needed for the Database() function */
  if (variable==NULL)
    {
      variable = (char *) malloc (MAX_QUERY_LENGTH+1);
      value = (char *) malloc (MAX_QUERY_LENGTH+1);
      tbuffer = (char *) malloc ((MAX_QUERY_LENGTH*2)+1);
      pbuffer = (char *) malloc ((sizeof(struct pcap_pkthdr)*2)+1);
    }
  if ( (variable==NULL) || (value==NULL) || (tbuffer==NULL) || (pbuffer==NULL))
    {
      Printf("Database: problem to allocate memory -> Exiting.\n");
      return(1);
    }

  /* now we allocate the memory for the util functions */
  util_alloc();

  /* Connect to database */
  if (Connect(data))
    {
      return(1);
    }

#ifdef DEBUG_DB
  Printf("database> After connect.\n");
#endif

  /* Get DB Version, needed to check if last_cid has to be set! */
  data->DBschema_version = CheckDBVersion(data);

#ifdef DEBUG_DB
  Printf("Database> schema version = %d\n", data->DBschema_version);
#endif

  if (data->DBschema_version > 105 )
    {
      snprintf(insert_ptr, MAX_QUERY_LENGTH, 
	       "INSERT INTO sensor (hostname, encoding, last_cid) "
	       "VALUES ('%s','%u','0')", 
	       data->sensor_name,  data->encoding);
    }
  else
    {
      snprintf(insert_ptr, MAX_QUERY_LENGTH, 
	       "INSERT INTO sensor (hostname, encoding) "
	       "VALUES ('%s','%u')", 
	       data->sensor_name,  data->encoding);
    }
  snprintf(select_ptr, MAX_QUERY_LENGTH, 
	   "SELECT sid FROM sensor WHERE hostname = '%s' "
	   "AND encoding = '%u' AND filter IS NULL",
	   data->sensor_name,  data->encoding);

#ifdef MYSQL_DB
  if ( (data->DBtype==DB_MYSQL) && data->transaction) 
      Insert("SET AUTOCOMMIT=0",data);
#endif

   if (data->transaction) DBsuccess=Insert("BEGIN",data);

#ifdef DEBUG_DB
  Printf("database> Select: <%s>\n",select_ptr);
#endif
  data->sid = Select(select_ptr,data);

#ifdef DEBUG_DB
  Printf("database> sid: <%s>\n",select_ptr);
#endif
  if(data->sid == 0)
    {
      /* This sensor is not in the database so we add him */
      result=Insert(insert_ptr,data);
      if (data->transaction && DBsuccess) DBsuccess=result;
#ifdef DEBUG_DB
      if (result==0)
	{
	  Printf("database> Insert failed: <%s>\n",insert_ptr);
	}
#endif

      data->sid = Select(select_ptr,data);
      if(data->sid == 0)
        {
	  Printf("Database> Problem obtaining SENSOR ID (sid) from "
		 "%s->sensor\n", data->dbname);
	  /* Should we exit here? I think we have a strange problem */
          /* with the database when this happens at this point...   */
	  Disconnect(data);
	  return(1);
        }
    }

#ifdef DEBUG_DB
  Printf("Database>     sensor id = %u\n", data->sid);
#endif

  /* Set the cid by getting the max value off the database */
  snprintf(select_ptr, MAX_QUERY_LENGTH, 
	   "SELECT MAX(cid) FROM event WHERE sid = '%u'",
          data->sid);
  data->cid=Select(select_ptr,data);

    /* Get the versioning information for the DB schema */

  if ( data->DBschema_version < 104 )
    {
      Printf("Database> The underlying database seems to be running"
	     " an older version of the DB schema.\n"
	     "          Please re-run the DB creation script (create_mysql)\n"
	     "          located in the contrib\\ directory.\n");
      if (data->transaction) Insert("ROLLBACK",data);
      Disconnect(data);
      return(2);
    }
  else if ( data->DBschema_version < LATEST_DB_SCHEMA_VERSION )
    {                
      Printf("Database> The database is using an older"
	     " version of the DB schema\n");
    }
  else if (data->DBschema_version > 105 )
    {
        /* Set the cid by getting the max value off the database */
      snprintf(select_ptr, MAX_QUERY_LENGTH, 
	       "SELECT last_cid FROM sensor WHERE sid = '%u'",
	       data->sid);
      cid = Select(select_ptr,data);      
      if ( data->cid > cid ) 
	{
	  /* on last exit the last_cid field was not updated!*/
	  snprintf(insert_ptr, MAX_QUERY_LENGTH, 
		   "UPDATE sensor SET last_cid='%u' WHERE sid='%u'", 
		   cid, data->sid);
	  result=Insert(insert_ptr,data);
          if (data->transaction && DBsuccess) DBsuccess=result;
#ifdef DEBUG_DB
	  if (result==0)
	    {
	      Printf("database> Insert failed: <%s>\n",insert_ptr);
	    }
#endif
	} 
      else
	{
	  /* Actualize data->cid to value of last_cid */
	  data->cid = cid;
	}

      /* Check for column data_header in table data */
      if (data->FullPayload)
        {
	  snprintf(select_ptr,MAX_QUERY_LENGTH,"SELECT full_payload FROM schema"); 
	  if (Select(select_ptr,data) == 0)
            {
	      /* Database is not able to store full payload.
               * Print a warning message and a hint how to
               * activate it. Additionally we have to disable
               * the feature and go on (or should we exit?)
	       */
	      data->FullPayload=0;
	      Printf("The database is not able to store the full "
		     "payload of the alerts although it is requested.\n");
	      Printf("To allow it, please extend the database with the "
		     "following SQL statements:\n\n");
              /* Needed for full header of the network packet */
	      Printf("ALTER TABLE data ADD COLUMN data_header TEXT;\n");
              /* pcap header, containing timestamp and length */
	      Printf("ALTER TABLE data ADD COLUMN pcap_header TEXT;\n");
              Printf("ALTER TABLE schema ADD COLUMN full_payload SMALLINT;\n");
              Printf("UPDATE schema SET full_payload=1;\n\n");
	    }
	} /* FullPayload */
    }

  /* Increment by one: this is the next cid! */
  data->cid++;
  if (data->transaction)
    {
      if (DBsuccess) 
	{ 
	  /* all inserts succeeded so we can commit */
	  result=Insert("COMMIT",data);
	}
      else
	{
	  /* at least one insert failed */
	  result=Insert("ROLLBACK",data);
	  /* With DatabaseInit only the sensor is added to the database */
	  /* if this already fails it makes no sense to continue...    */
	  Printf("DatabaseInit> There was an error with the database\n");
	  Disconnect(data);
	  return(1);
	}
    }
  return(0);
}


/*
 * Function: Database(AlertPacket *, char * msg, void *arg)
 *
 * Purpose: Insert data into the database
 *
 * Arguments: p   => pointer to the current Alertpacket data struct 
 *            msg => pointer to the signature (alert) message
 *            arg => pointer to the data associated with the database
 *            event => pointer to the event structure
 *            Packet_ptr => pointer to the sniffed packet
 *
 * Returns: void function
 *
 */

int 
Database(AlertpktDB *p, char *msg, void *arg, Event *event, void *Packet_ptr)
{
  DatabaseData *data = (DatabaseData *)arg;
  char *time, *tmp;
  char *sig_msg;
  int i, dsize, hsize, len, ins;
  unsigned int sig_id=0;
  int NewRule;  
  int result, DBsuccess=0;
  
#ifdef DEBUG_DB
  Printf("Database> AlertpktDB: %p\n",p);
#ifndef GENUA
  Printf("Database> Message   : %s\n",msg);
#endif
  Printf("Database> DBdata    : %p\n",arg);
  Printf("Database> Packet_ptr: %p\n",Packet_ptr);
#endif

  if(msg == NULL)
    {
      msg = "";
    }

  /*** Build the query for the Event Table ***/
  if(!(p->val & NOPACKET_STRUCT))
    {
      time = GetTimeStamp((struct timeval *)&p->pkth.ts, data->tz);
    }
  else
    {
      /* Ok, there is no sniffed packet with timestamp so we */
      /* use the actual time. This is not very accurate here */
      time = GetCurrentTimeStamp();
    }

  if (event->sig_generator != GENERATOR_SNORT_ENGINE) /*(event->sig_id < 101) */
    {
       /* This alert is created by a preprocessor: We replace   */
       /* sig_rev with the generator id to separate the sig_sid */
       /* of the different sources. (sig_sid are only unique    */
       /* within each preprocessor! And yes: It is a hack...)   */

	event->sig_rev = event->sig_generator;
    }           

  if (data->transaction) DBsuccess=Insert("BEGIN",data);

  NewRule=1; /* First we don't trust anything! */

  if (data->DBtrust)
    {
       /* We trust now that the sig_id is already included in the */
       /* database. We can therefore skip all the next lines... */    

      snprintf(select_ptr, MAX_QUERY_LENGTH,
	       "SELECT sig_id FROM signature WHERE sig_priority = '%u' AND "
	       " sig_rev = '%u' AND sig_sid = '%u' ",
	       event->priority, event->sig_rev, event->sig_id);

      sig_id=Select(select_ptr,data);
#ifdef DEBUG_DB
      Printf("Trusting Database, sig_id: %i\n",sig_id);
      Printf("Database> %s\n",select_ptr);
#endif
      if (sig_id > 0 ) 
	 {
	     /* Yes, we can trust the DB... */
             NewRule=0;
          }
       else 
	{
           Printf("Disabling DBtrust modus for this alert:\n");
	   Printf("[%i] %s\n\n",event->sig_id,msg);
        }
    }
  if (NewRule)
    {
#ifdef GENUA
      Printf("Houston, we have a problem!\n");
      Printf("Here is a rule which is not in the database!\n");
      exit(1);
#else
      /* 
       * Determine the ID # of the signature of this alert 
       *
       * Note: Here we still use the old syntax of spo_database.c
       *       Therfore values which are zero are assumed to be
       *       not set so we ask for NULL values. I think a better
       *       approach is to ask for zeroes instead of NULL values
       *       like we do in DBTrust.
       *
       * Note: We don't count check here for the priority! So if
       *       there is a rule with several different priorities
       *       we still take the first sig_id of the database!
       *       This is another point which we should fix? But
       *       DBTrust will do it correctly.
       */

      sig_msg=snort_escape_string(msg,data);

      if ( event->sig_rev == 0 ) 
        {
          if( event->sig_id == 0) 
	    {
	      
	      snprintf(select_ptr, MAX_QUERY_LENGTH, 
		       "SELECT sig_id FROM signature WHERE sig_name = '%s' AND"
		       " sig_rev is NULL AND sig_sid is NULL ", sig_msg);
	    }
          else 
	    {
	      snprintf(select_ptr, MAX_QUERY_LENGTH, 
		       "SELECT sig_id FROM signature WHERE sig_name = '%s' AND"
		       " sig_rev is NULL AND sig_sid = %u ", 
		       sig_msg, event->sig_id);
	    }
        }
      else
        {
          if( event->sig_id == 0)
	    {
	      snprintf(select_ptr, MAX_QUERY_LENGTH,
		       "SELECT sig_id FROM signature WHERE sig_name = '%s' AND"
		       " sig_rev = %u AND sig_sid is NULL ",
		       sig_msg, event->sig_rev);
	    }
          else
	    {
	      snprintf(select_ptr, MAX_QUERY_LENGTH,
		       "SELECT sig_id FROM signature WHERE sig_name = '%s' AND"
		       " sig_rev = %u AND sig_sid = %u ",
		       sig_msg, event->sig_rev, event->sig_id);
	    }
        }
    
      sig_id = Select(select_ptr, data);
    
#ifdef DEBUG_DB
  Printf("Database> sig_id: %i\n",sig_id);
#endif 

      /* We should have another method to insert the complete rule into */
      /* the database. Here we can only insert minimalistic information */
      /* At this point we have no OTN to fill in the missing values     */
      if(sig_id == 0)
        {
	  bzero(variable,MAX_QUERY_LENGTH+1);
	  bzero(value,MAX_QUERY_LENGTH+1);
      
          strcpy(variable, "sig_name");
          snprintf(value, MAX_QUERY_LENGTH, "'%s'", sig_msg);
      
	  len=strlen(value);

	  /* Maybe we should include all values even if they are zero?  */
          /* Here we build only an INSERT string consisting of non-zero */
          /* values. So we will end up with some NULL values in a rule. */

          if ( event->priority > 0 )
	    {
	      strcat(variable, ",sig_priority");
	      snprintf( (value+len), (MAX_QUERY_LENGTH-len), 
			",%u", event->priority);
	      len=strlen(value);
	    }
      
          if ( event->sig_rev > 0 )
	    {
	      strcat(variable, ",sig_rev");
	      snprintf((value+len), (MAX_QUERY_LENGTH-len), ",%u", 
		       event->sig_rev);
	      len=strlen(value);
	    }
          
          if ( event->sig_id > 0 )
	    {
	      strcat(variable, ",sig_sid");
	      snprintf((value+len), (MAX_QUERY_LENGTH-len), ",%u", 
		       event->sig_id);
	      len=strlen(value);
	    }
          
          if ( event->classification > 0 )
	    {
	      strcat(variable, ",sig_class_id");
	      snprintf((value+len), (MAX_QUERY_LENGTH-len), ",%u", 
		       event->classification);
	    }

          snprintf(insert_ptr, MAX_QUERY_LENGTH,
	           "INSERT INTO signature (%s) VALUES (%s)",
	           variable, value);
      
          result=Insert(insert_ptr,data);
#ifdef DEBUG_DB
	  if (result==0)
	    {
	      Printf("database> Insert failed: <%s>\n",insert_ptr);
	    }
#endif

	  /* if we use transaction and up to now we succeeded
	   * with inserts then we set DBsuccess with the result
	   * of the insert. On success it remains 1, otherwise
	   * it is set to 0. 
           */

	  if (DBsuccess) DBsuccess=result;

          sig_id = Select(select_ptr,data);
          if(sig_id == 0)
	    {
	      /* We still have no sig_id: This should not happen! */
	      Printf("Database> Problem inserting a new signature '%s'\n", msg);
              /* As stated above: Should we exit here? */
	    }
        } /* end of first: if (sig_id==0) */

      /* If we have a method to fill in the above values via another */
      /* method we can remove this code up to here... see -> DBtrust */
#endif /* GENUA */
    } /* end of NewRule */

  snprintf(insert_ptr, MAX_QUERY_LENGTH,
	   "INSERT INTO event (sid,cid,signature,timestamp) VALUES "
	   "('%u', '%u', '%u', '%s')",
	   data->sid, data->cid, sig_id, time);
  result=Insert(insert_ptr,data);
  if (DBsuccess) DBsuccess=result;
#ifdef DEBUG_DB
	  if (result==0)
	    {
	      Printf("database> Insert failed: <%s>\n",insert_ptr);
	    }
#endif

#ifdef DEBUG_DB
  Printf("Database> p: %p\n",p);
  Printf("Database> p->val: %i\n",p->val);
#endif
  if(!(p->val & NOPACKET_STRUCT)) 
    {
      /* We do not log fragments! They are assumed to be handled */
      /* by the fragment reassembly pre-processor                */
      if((!p->frag_flag) && p->iph.ip_verhl ) 
	{
	  /*
	   *    >>> ICMP <<<
	   */
	  if(p->iph.ip_proto == IPPROTO_ICMP)
	    {
	      /*** Build a query for the ICMP Header ***/
	      if(p->icmp_ext)
		{
		  snprintf(insert_ptr, MAX_QUERY_LENGTH, 
			   "INSERT INTO icmphdr (sid, cid, icmp_type, "
			   "icmp_code, icmp_csum, icmp_id, icmp_seq) "
			   "VALUES ('%u','%u','%u','%u','%u','%u','%u')",
			   data->sid, data->cid, p->icmp_type, p->icmp_code,
			   ntohs(p->icmp_csum), ntohs(p->icmp_extid), 
			   ntohs(p->icmp_extseqno));
		}
	      else
		{
		  snprintf(insert_ptr, MAX_QUERY_LENGTH, 
			   "INSERT INTO icmphdr (sid, cid, icmp_type, "
			   "icmp_code, icmp_csum) "
			   "VALUES ('%u','%u','%u','%u','%u')",
			   data->sid, data->cid, p->icmp_type, p->icmp_code,
			   ntohs(p->icmp_csum));
		}
          result=Insert(insert_ptr,data);
	  if (DBsuccess) DBsuccess=result;
#ifdef DEBUG_DB
	  Printf("Database> ICMP: %s\n",insert_ptr); 
	  if (result==0)
	    {
	      Printf("database> Insert failed: <%s>\n",insert_ptr);
	    }
#endif
	    }
	  else if (p->iph.ip_proto == IPPROTO_TCP ) /* && p->tcph.th_seq) */
	    /*
	     *              >>> TCP <<<
	     */
            {
	      /*** Build a query for the TCP Header ***/
	      snprintf(insert_ptr, MAX_QUERY_LENGTH, 
		       "INSERT INTO tcphdr "
		       "(sid, cid, tcp_sport, tcp_dport, tcp_seq,"
		       " tcp_ack, tcp_off, tcp_res, tcp_flags, tcp_win,"
		       " tcp_csum, tcp_urp) "
		       "VALUES ('%u','%u','%u','%u','%lu','%lu','%u',"
		       "'%u','%u','%u','%u','%u')",
		       data->sid, data->cid, ntohs(p->tcph.th_sport), 
		       ntohs(p->tcph.th_dport), (u_long)ntohl(p->tcph.th_seq),
		       (u_long)ntohl(p->tcph.th_ack), TCP_OFFSET_DG(p->tcph), 
		       TCP_X2_DG(p->tcph), p->tcph.th_flags, 
		       ntohs(p->tcph.th_win), ntohs(p->tcph.th_sum),
		       ntohs(p->tcph.th_urp));
	      result=Insert(insert_ptr,data);
	      if (DBsuccess) DBsuccess=result;
#ifdef DEBUG_DB
	      Printf("Database> TCP: %s\n",insert_ptr); 
	      if (result==0)
		{
		  Printf("database> Insert failed: <%s>\n",insert_ptr);
		}
#endif
                    /*** Build the query for TCP Options ***/
	      for(i=0; i < (int)(p->tcp_option_count); i++)
		{
		  if((data->encoding == ENCODING_HEX) || 
		     (data->encoding == ENCODING_ASCII))
		    {
		      tmp = Fasthex(p->tcp_options[i].offset+Packet_ptr, 
				    p->tcp_options[i].len); 
		    }
		  else
		    {
		      tmp = Base64(p->tcp_options[i].offset+Packet_ptr, 
				   p->tcp_options[i].len); 
		    }
		  snprintf(insert_ptr, MAX_QUERY_LENGTH, 
			   "INSERT INTO opt "
			   "(sid, cid, optid, opt_proto, opt_code, opt_len, "
			   "opt_data) "
			   "VALUES ('%u','%u','%u','%u','%u','%u','%s')",
			   data->sid, data->cid, i, 6, p->tcp_options[i].code,
			   p->tcp_options[i].len, tmp); 
		  result=Insert(insert_ptr,data);
		  if (DBsuccess) DBsuccess=result;
#ifdef DEBUG_DB
		  Printf("Database> TCP: %s\n",insert_ptr); 
		  if (result==0)
		    {
		      Printf("database> Insert failed: <%s>\n",insert_ptr);
		    }
#endif
		}
	    }
	  else if(p->iph.ip_proto == IPPROTO_UDP && p->udph.uh_len)
	    {
	      /*
	       *             >>> UDP <<<
	       */
	      /*** Build the query for the UDP Header ***/
	      
	      snprintf(insert_ptr, MAX_QUERY_LENGTH,
		       "INSERT INTO udphdr "
		       "(sid, cid, udp_sport, udp_dport, udp_len, udp_csum) "
		       "VALUES ('%u', '%u', '%u', '%u', '%u', '%u')",
		       data->sid, data->cid, ntohs(p->udph.uh_sport), 
		       ntohs(p->udph.uh_dport), ntohs(p->udph.uh_len),
		       ntohs(p->udph.uh_chk));
	      result=Insert(insert_ptr,data);
	      if (DBsuccess) DBsuccess=result;
#ifdef DEBUG_DB
	      Printf("Database> UDP: %s\n",insert_ptr); 
	      if (result==0)
		{
		  Printf("database> Insert failed: <%s>\n",insert_ptr);
		}
#endif
	    }
	}   
      
      /*** Build the query for the IP Header ***/
      if ( p->iph.ip_verhl )
	{
	  /*
	   *    >>> IP <<<
	   */
	  
	  snprintf(insert_ptr, MAX_QUERY_LENGTH, 
		   "INSERT INTO iphdr "
		   "(sid, cid, ip_src, ip_dst, ip_ver,"
		   "ip_hlen, ip_tos, ip_len, ip_id, ip_flags, ip_off,"
		   "ip_ttl, ip_proto, ip_csum) "
		   "VALUES ('%u','%u','%lu',"
		   "'%lu','%u',"
		   "'%u','%u','%u','%u','%u','%u',"
		   "'%u','%u','%u')",
		   data->sid, data->cid, (u_long)ntohl(p->iph.ip_src.s_addr), 
		   (u_long)ntohl(p->iph.ip_dst.s_addr), 
		   IP_VER_DG(p->iph), IP_HLEN_DG(p->iph), 
		   p->iph.ip_tos, ntohs(p->iph.ip_len), ntohs(p->iph.ip_id), 
		   p->frag_flag, ntohs(p->frag_offset), p->iph.ip_ttl, 
		   p->iph.ip_proto, ntohs(p->iph.ip_csum));
	  result=Insert(insert_ptr,data);
	  if (DBsuccess) DBsuccess=result;
	}
#ifdef DEBUG_DB
   Printf("Database> IP: %s\n",insert_ptr); 
   if (result==0)
     {
       Printf("database> Insert failed: <%s>\n",insert_ptr);
     }
#endif

      
      /*** Build querys for the IP Options ***/
      for(i=0 ; i < (int)(p->ip_option_count); i++)
	{
	  if(&p->ip_options[i])
	    {
	      if((data->encoding == ENCODING_HEX) || 
		 (data->encoding == ENCODING_ASCII))
		{
		  tmp = Fasthex(p->ip_options[i].offset+Packet_ptr, 
				p->ip_options[i].len); 
		}
	      else
		{
		  tmp = Base64(p->ip_options[i].offset+Packet_ptr, 
			       p->ip_options[i].len); 
		}
	      snprintf(insert_ptr, MAX_QUERY_LENGTH, 
		       "INSERT INTO opt "
		       "(sid,cid,optid,opt_proto,opt_code,opt_len,opt_data) "
		       "VALUES ('%u','%u','%u','%u','%u','%u','%s')",
		       data->sid, data->cid, i, 0, p->ip_options[i].code,
		       p->ip_options[i].len, tmp); 
	      result=Insert(insert_ptr,data);
	      if (DBsuccess) DBsuccess=result;
#ifdef DEBUG_DB
   Printf("Database> IP: %s\n",insert_ptr); 
   if (result==0)
     {
       Printf("database> Insert failed: <%s>\n",insert_ptr);
     }
#endif

	    }
	}
    }
#ifdef DEBUG_DB  
  Printf("Database> After header: %s\n",insert_ptr);
#endif
        /*** Build query for the payload ***/
  /* dsize: size of payload without headers, p->pkth.caplen is     */
  /*        total of captured packet, p->data is pointer to the   */
  /*        payload and Packet_ptr is the pointer to the beginning */
  /*        of the captured packet inclusive headers!              */
  /* dsize = (int) (Packet_ptr - p->data) + p->pkth.caplen; */
  /* maybe the captured len (caplen) is longer than the real length */
  /* Some UUUU's from 100 MBit? */

  dsize = (int) (Packet_ptr - p->data) + p->pkth.len;
  hsize = (int) (p->data) - (int) (Packet_ptr); 

  if (hsize > MAXBUFFSIZE) 
    {
      Printf("Error: Header size too large -> reducing to MAXBUFFSIZE\n");
      hsize=MAXBUFFSIZE;
    }

#ifdef DEBUG_DB
  Printf("Database> DataSize: %i, HeaderSize: %i, Packet_ptr: %p, data: %p\n",
	 dsize, hsize, Packet_ptr,p->data);
  Printf("Database> packet length: %u, captured length: %u\n",
          p->pkth.len, p->pkth.caplen);
#endif

  if (hsize)
    {
      if(data->encoding == ENCODING_BASE64)
	{
	      tmp  = Base64((u_char *)Packet_ptr, hsize);
              strcpy(tbuffer,tmp);
              tmp  = Base64((u_char *)&p->pkth,sizeof(struct pcap_pkthdr));
              strcpy(pbuffer,tmp);
	}
      else
	{
	      tmp  = Fasthex((u_char *) Packet_ptr, hsize);
              strcpy(tbuffer,tmp);
              tmp  = Fasthex((u_char *)&p->pkth,sizeof(struct pcap_pkthdr));
#ifdef DEBUG_DB
	      printf("pcap_pkthdr: %i\n",sizeof(struct pcap_pkthdr));
	      printf("pcap-len: %i\n",p->pkth.caplen);
#endif
              strcpy(pbuffer,tmp);
	}
    }

  if (dsize)
    {
      if(data->encoding == ENCODING_BASE64)
	{
	      tmp  = Base64((u_char *)p->data, dsize);
	}
      else
	{
	  if(data->encoding == ENCODING_ASCII)
	    {
	      tmp = Ascii((u_char *)p->data, dsize);
	    }
	  else
	    {
	      tmp  = Fasthex((u_char *)p->data, dsize);
	    }
	}
    }
  
  ins=0;

  if (hsize && dsize)
    {
      ins=1;
      if (data->FullPayload)
	{
	  snprintf(insert_ptr, MAX_QUERY_LENGTH - 3, 
		   "INSERT INTO data "
		   "(sid,cid,data_payload,pcap_header,data_header) "
		   "VALUES ('%u','%u','%s','%s','%s",
		   data->sid, data->cid, tmp, pbuffer, tbuffer);
	}
      else
	{
	  snprintf(insert_ptr, MAX_QUERY_LENGTH - 3, 
		   "INSERT INTO data "
		   "(sid,cid,data_payload) "
		   "VALUES ('%u','%u','%s",
		   data->sid, data->cid, tmp);
	}
    }
  else if ( hsize && (dsize==0) )
    {
      if (data->FullPayload)
	{
	  ins=1;
	  snprintf(insert_ptr, MAX_QUERY_LENGTH - 3, 
		   "INSERT INTO data "
		   "(sid,cid,pcap_header,data_header) "
		   "VALUES ('%u','%u','%s','%s",
		   data->sid, data->cid, pbuffer, tbuffer);
	}
    }

  /* Do we have something to insert? */
  if (ins)
    {
      strcat(insert_ptr, "')");
      result=Insert(insert_ptr,data);
      if (DBsuccess) DBsuccess=result;
#ifdef DEBUG_DB
      Printf("Database> data: %s\n",insert_ptr); 
      if (result==0)
	{
	  Printf("database> Insert failed: <%s>\n",insert_ptr);
	}
#endif
    }

  data->cid++;
#ifdef DEBUG_DB
  Printf("Database> Return, next cid: %i.\n",data->cid);
#endif
  if (data->transaction)
    {
      if (DBsuccess) 
	{ 
	  /* all inserts succeeded so we can commit */
	  result=Insert("COMMIT",data);
	}
      else
	{
	  /* at least one insert failed */
	  result=Insert("ROLLBACK",data);
	}
    }
  return(DBsuccess);
}

/* Some of the code in this function is from the 
   mysql_real_escape_string() function distributed with mysql.

   Those portions of this function remain
   Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB

   We needed a more general case that was not MySQL specific so there
   were small modifications made to the mysql_real_escape_string() 
   function. */

char * snort_escape_string(char * from, DatabaseData * data)
{
  /* the only string we have to escape is the altermsg */
  static char to_start[ALERTMSG_LENGTH*2+1];
  char * to;
  char * end; 
  int from_length;

  from_length = (int)strlen(from);

  to = to_start;
  {
    for(end=from+from_length; from != end; from++)
      {
        switch(*from)
        {
          case 0:             /* Must be escaped for 'mysql' */
            *to++= '\\';
            *to++= '0';
            break;
          case '\n':              /* Must be escaped for logs */
            *to++= '\\';
            *to++= 'n';
            break;
          case '\r':
            *to++= '\\';
            *to++= 'r';
            break;
          case '\\':
            *to++= '\\';
            *to++= '\\';
            break;
          case '\'':
            *to++= '\\';
            *to++= '\'';
            break;
          case '"':               /* Better safe than sorry */
            *to++= '\\';
            *to++= '"';
            break;
          case '\032':            /* This gives problems on Win32 */
            *to++= '\\';
            *to++= 'Z';
            break;
          default:
            *to++= *from; 
        }
      }
  }
  *to=0;
  return(char *)to_start;
}


/* Function: CheckDBVersion(DatabaseData * data)
 *
 * Purpose: To determine the version number of the underlying DB schema
 *
 * Arguments: database information
 *
 * Returns: version number of the schema
 */

int 
CheckDBVersion(DatabaseData * data)
{
  int schema_version;

  snprintf(select_ptr, MAX_QUERY_LENGTH,  "SELECT vseq FROM schema");

  schema_version = Select(select_ptr,data);

  return schema_version;
}


/* Function: Insert(char * query, DatabaseData * data)
 *
 * Purpose: Database independent function for SQL inserts
 * 
 * Arguments: query (An SQL insert)
 *
 * Returns: 1 if successful, 0 if fail
 */

int 
Insert(char * query, DatabaseData * data)
{
    int result = 0;

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(!(mysql_query(data->m_sock,query)))
        {
            result = 1;
        }
        else
        {
            if(mysql_errno(data->m_sock))
            {
              Printf("database: mysql_error: %s\nSQL=%s\n", 
                           mysql_error(data->m_sock), query);
            }
        }
      }
#endif /* MYSQL */

#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_result = PQexec(data->p_sock,query);
	if (PQresultStatus(data->p_result) == PGRES_COMMAND_OK)
	    {
	      result = 1;
	    }
	else
	  {
	    if (PQerrorMessage(data->p_sock))
	      Printf("database: psql_error: %s\nSQL=%s\n",
		   PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }
#endif /* POSTGRES */
#ifdef DEBUG_DB
    if(result)
        Printf("Database> (%s) executed\n", query);
    else
        Printf("Database> (%s) failed\n", query);
#endif

    return result;
}


/* Function: Select(char * query, DatabaeData * data)
 *
 * Purpose: Database independent function for SQL selects that 
 *          return a non zero int
 * 
 * Arguments: query (An SQL insert)
 *
 * Returns: result of query if successful, 0 if fail
 */

int 
Select(char * query, DatabaseData * data)
{
    int result = 0;

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(mysql_query(data->m_sock,query))
	  {
            result = 0;
	  }
        else
	  {
            if(!(data->m_result = mysql_use_result(data->m_sock)))
	      {
                result = 0;
	      }
            else
	      {
                if((data->m_row = mysql_fetch_row(data->m_result)))
		  {
                    if(data->m_row[0] != NULL)
		      {
                        result = atoi(data->m_row[0]);
		      }
		  }
	      }
            mysql_free_result(data->m_result);
	  }
        if(!result)
	  {
            if(mysql_errno(data->m_sock))
	      {
                Printf("database: mysql_error: %s\n", 
		       mysql_error(data->m_sock));
	      }
	  }
      }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_result = PQexec(data->p_sock,query);
	if ( PQresultStatus(data->p_result) == PGRES_TUPLES_OK)
	  {
	    if (PQntuples(data->p_result))
	      result = atoi(PQgetvalue(data->p_result,0,0));
#ifdef DEBUG_DB
	Printf("Status: %i\n",result);
#endif
	  }
	else
	  {
	    result=0;
	    if (PQerrorMessage(data->p_sock))
	      Printf("database: psql_error: %s\nSQL=%s\n",
		     PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }
#endif /* POSTGRES */
#ifdef DEBUG_DB
    if(result)
        Printf("Database> (%s) returned %u\n", query, result);
    else
        Printf("Database> (%s) failed\n", query);
#endif

    return result;
}


/* Function: Connect(DatabaseData * data)
 *
 * Purpose: Database independent function to initiate a database 
 *          connection
 */

int 
Connect(DatabaseData * data)
{
#ifdef MYSQL_DB  
  int x; 
  if ( data->DBtype == DB_MYSQL) 
    {
      data->m_sock = mysql_init(NULL);
      if(data->m_sock == NULL)
	{
	  Printf("database: Connection to database '%s' failed\n", 
		 data->dbname);
	  return(2);
	}
      x = 0;
      if( mysql_real_connect(data->m_sock,"localhost", data->user,
	       data->password, data->dbname, x, data->socket_name, 0) == 0)
	{
	  if(mysql_errno(data->m_sock))
	    {
	      Printf("database: mysql_error: %s\n", mysql_error(data->m_sock));
	      return(1);
	    }
	  Printf("database: Failed to logon to database '%s'\n", data->dbname);
	  return(1);
	}
    }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_sock=PQsetdbLogin(NULL, data->socket_name ,NULL,NULL,
				  data->dbname, data->user, data->password);
	if (PQstatus(data->p_sock) == CONNECTION_BAD)
	  {
	    PQfinish(data->p_sock);
	    if (PQerrorMessage(data->p_sock))
	      Printf("database: psql_error: %s\n",
		     PQerrorMessage(data->p_sock));
	    Printf("database: Failed to logon to database '%s'\n",
		   data->dbname);
	    return(1);
	  }
      }
#endif /* POSTGRES */
    return(0);
}

/* Function: Disconnect(DatabaseData * data)
 *
 * Purpose: Close connection to database and free 
 *          allocated memory.
 */

void 
Disconnect(DatabaseData * data)
{
  static int called=0;

  if (called) return;
  called++;

#ifdef DEBUG_DB
  Printf("Database> Closing connection to database \"%s\"\n", data->dbname);
#endif
  if (data)
    {
      if ( data->DBschema_version > 105 )
	{
	  /* on exit update last_cid to last cid!*/
	  snprintf(insert_ptr, MAX_QUERY_LENGTH, 
		   "UPDATE sensor SET last_cid='%u' WHERE sid='%u'",
		   data->cid, data->sid);
	  Insert(insert_ptr,data);
	}
#ifdef MYSQL_DB
      if((data->DBtype==DB_MYSQL) && data->m_sock ) 
	mysql_close(data->m_sock);
#endif 
#ifdef POSTGRES
      if ( (data->DBtype==DB_POSTGRES) && data->p_sock) 
	PQfinish(data->p_sock);
#endif
    }

  /* to be kindly we free some memory... (but exit() will do it too... */
  free(insert_ptr);
  free(select_ptr);
}
